(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Arg;;

let verbose = ref 0 
let inchan = ref stdin 
let outchan = ref stdout 
let inchanset = ref false
let outchanset = ref false

let usagestring = "Usage : SXC [options] [inputfile]";;

let rec args = 
  [ "-v", Unit verbose_on, "Turn verbose mode up (can be scary)";
    "-o", String set_outchan, "Output to specified file";
    "--help", Unit print_usage, "Prints this help information" ]

and verbose_on () = incr verbose

and set_outchan s = 
  if not !outchanset then try
    outchan := open_out s ;
    outchanset := true
  with Sys_error _ -> prerr_string ("Unable to open file "^s^" for output\n"); exit 1 
  else
    (usage args "Only one output file please."; exit 1)

and others s = 
  if not !inchanset then try 
    inchan := open_in s;
    inchanset := true
  with Sys_error _ -> prerr_string ("Unable to open file "^s^" for input\n"); exit 1
  else
    (usage args "Only one input file please."; exit 1)

and print_usage () =
  usage args "Usage : SXC [options] [inputfile]";
  exit 0
;;

parse args others usagestring;;

let inchan = !inchan
let outchan = !outchan
let verbose = !verbose
