(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

type temp = Gen.temp
type label = Gen.label
      
type instr =
  | Oper of string * temp list * temp list * bool * label list
  | Move of temp * temp
  | Label of string * label 

type func =
{ 
    code : instr list; (* Annotated assembler code *)
    flow : flowinfo list; (* Flow information *)
} 

and flowinfo = {
    instr : instr;
    def : temp Smallset.set;  use : temp Smallset.set; 
    mutable live_in : temp Smallset.set;
    mutable live_out : temp Smallset.set;
    mutable succ : flowinfo list;
    mutable dead : bool;
  } ;;

type code = instr list
        
let nop = Oper ("", [], [], true, [])
let escape = '^'
    
exception Instruction of string * temp list * temp list
    
let digit string i =
  let n = Char.code string.[i] - Char.code '0' in
  if 0 <= n && n <= 3 then n else failwith "digit"
    
let substring s i j =  String.sub s i (j+1-i)
        
exception Format of instr

let rec string_instr namer string src dest i =
  if i < String.length string then
    begin try 
      let j = String.index_from string i escape in
      (substring string i (j-1))^
      let list =
        match string.[j+1] with
        | 'd' -> dest | 's' -> src
        | _ -> failwith "format" in
      let n = digit string (j+2) in
      (namer (List.nth list n))^(string_instr namer string src dest (j+3))
    with
      Not_found -> 
        (String.sub string i (String.length string -i))
    | Failure s ->
        raise (Instruction (s^":"^string, src, dest))
    end else ""
      
let indent = "    ";;
let to_string namer instr =
  try
    begin match instr with
    | Oper (s, src, dest, follow, labs) ->
        indent^(string_instr namer s src dest 0)
    | Label (s, l) ->
        s^":"
    | Move (src, dest) ->
        indent^(string_instr namer "mov ^d0, ^s0" [src] [dest] 0)
    end
  with
    x ->
      raise (Format instr)
        ;;

let print chan namer instr = 
  output_string chan (to_string namer instr)

let nformat = print stdout
let format = nformat Gen.namer

(*
let print_reg t = print_string (Gen.temp_string t); print_string " ";;
let print_reg_list l = print_string "\t# "; List.iter print_reg l;;
let print_sd src dest = print_reg_list src; print_reg_list dest;;
*)
