(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Code;;

exception BackendFailure of string;;

let funcs = ref [];;

let rec emit_func depth f =
  if not f.translated then
    begin
      if depth > f.depth then f.depth <- depth;
      if depth > 8 then failwith "Too many call levels (limited to 8, including main)";
      if f.translating then raise (BackendFailure "Recursive function call!");
      f.translating <- true;
      List.iter (emit_func (depth + 1)) f.usefuncs;
      let ass = (Ass.Oper ("", [], (List.flatten (f.statics::f.args)), true, []))
	::(Codegen.emit_code f.code) in ();
      let flow = Liveness.flow ass in
      if Args.verbose > 0 then begin
	print_string "\nPrinting initial flow graph.*************************************************\n";
	Liveness.print_function_flow flow;
	print_string "\nEnd of initial flow graph.*************************************************\n\n";
      end;
      let flow = Liveness.dead_code_killer Codegen.always_used_registers flow in
      let ass = Liveness.code_of_graph flow in
      f.uses <- (List.hd flow).Ass.live_out;
      f.defs <- List.flatten (List.map (function n -> n.Ass.def) (List.tl flow)); 
      funcs := { Ass.code = ass; Ass.flow = flow }::!funcs;
      f.translating <- false;
      f.translated <- true;
    end;;

let emit_program p =
  funcs := [];
  emit_func 0 p.main;
  let result = !funcs in
  funcs := [];
  result;;


  
