(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

type exp =
    Const of int list
  | Name of variable
  | Temp of Gen.temp list (* Gen.label *) (* Gen.label indique si var statique *)
  | Un of unop * exp
  | Bin of binop * exp * exp
  | Call of func * exp list
  | Ifte of exp * exp * exp * int

and stm =
  | Label of Gen.label
  | Seq of stm list
  | Exp of exp
  | Jump of Gen.label
  | Cjump of exp * bool * Gen.label * Gen.label
  | Return of func * exp

and binop = Plus | Minus | STimes | UTimes | USTimes | UDiv | SDiv | USDiv 
	    | UMod | SMod | USMod | SUMod
            | And | Or | Xor
            | SLt | SLe | SGt | SGe | Eq | Ne
            | ULt | ULe | UGt | UGe 
            | Assign | Comma | Concat | Assignbit of int
and unop = Neg | Not | Asr | Shr | Shl | Rol | Ror | Ro4 | UExtend of int | SExtend of int
  | Extract of int * int
  | Extractbit of int
  | NotZero | SAssign of Gen.temp

and code = stm list

and variable = {
  mutable vid : int;   (*identifiant unique dans tout le programme*)
  vname : int;         (* index dans la hashtable du lexer*)
  mutable typ : Pp.type_expr;
  static : bool;
  mutable temps : Gen.temp list;
}

and func =
{ 
    mutable code : code; (* Intermediate code *)
    name : Gen.label; (* Function start label *)
    args : Gen.temp list list; (* Temporaries for function's arguments *)
    result : Gen.temp list; (* Temporary where return value must be placed *)
    resulttype : Pp.type_expr; (* Result type if result != [] else it is set to Byte*)
    mutable depth : int;
    mutable statics : Gen.temp list; (* Temporaries that must be maintained outside of function *)
    mutable uses : Gen.temp list; (* Registers used by function *)
    mutable defs : Gen.temp list; (* Registers defined by function *)
    mutable usefuncs : func list; (* Functions called by this function *)
    mutable translating : bool; (* Indicates that translation to assembly has started *)
    mutable translated : bool; (* Indicates that translation to assembly has finished *)
} 

and program =
{ 
    main : func;
    funcs : func list;
}

let to_string = function
    Label l -> "Label "^(Gen.label_string l)
  | Seq _ -> "      Sequence"
  | Exp _ -> "      expression"
  | Jump l -> "      Jump "^(Gen.label_string l)
  | Cjump (_, _, l1, l2) -> "      CJump "^(Gen.label_string l1)^" "^(Gen.label_string l2)
  | Return _ -> "      Return"

let output chan stm =
  output_string chan (to_string stm)

let output_list chan stms =
  List.iter (fun stm -> output chan stm; output_char chan '\n') stms
