(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

type instr = 
    Oper of string * int
  | Label of Gen.label
  | Jcond of string * Gen.label * Gen.label
  | Jmp of Gen.label
  | Call of Gen.label
  | Bank of int
  | Ret
;;

let instr_string = function
    Label l -> Gen.label_string l
  | Oper (s, _) -> s
  | Jmp l -> "    jmp "^(Gen.label_string l)
  | Call l -> "    call "^(Gen.label_string l)
  | Jcond (s, l1, l2) -> 
      "    snb "^s^" ; as one jcond"^
      "\n    jmp "^(Gen.label_string l1)^
      "\n    jmp "^(Gen.label_string l2)
  | Ret -> "    ret"
  | Bank i -> "    bank #"^(string_of_int i)
;;

let instr_words = function
    Label _ -> 0
  | Oper (_, i) -> i
  | Jmp _ -> 1
  | Call l -> 1
  | Jcond _ -> 3
  | Ret -> 1
  | Bank i -> 1
;;
