(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

let last_label = ref (-1)
let last_temp = ref (-1)

type label = {id : int; name : string}
and temp = int

let temp_names = Hashtbl.create 101;;

let new_temp() = incr last_temp; !last_temp

let name_temp t v = 
  Hashtbl.add temp_names t v;;

let registers =
  Array.init 300 (fun i -> new_temp ());;
(*let steal_reg i tt = Hashtbl.add temptypes i tt; i;;*)

let label_names = Hashtbl.create 101
let create_label id name =
  let l = { id = id; name = name; } in
  Hashtbl.add label_names name l; l

let new_label() =
  incr last_label;
  create_label !last_label ("L"^(string_of_int !last_label))

let names = Hashtbl.create 3
let prefixed_label s =
  let n, sn = 
    try let n = succ (Hashtbl.find names s) in n, s^"_"^(string_of_int n)
    with Not_found -> 0, s in
  incr last_label;
  let l = create_label !last_label sn in
  Hashtbl.add names s n; l
;;

let named_label s =
  try let _ = Hashtbl.find names s in failwith "named_label"
  with Not_found -> prefixed_label s
  ;;

let label_of_name s =
  try Hashtbl.find label_names s
  with Not_found -> raise (Failure ("Label not found "^s))
  ;;
  
(* pour imprimer les instructions et pour la mise au point *)
let label_string l = l.name
(*let temp_string t = match temp_type t with
  Byte -> "t"^(string_of_int t)
| Bool -> "tb"^(string_of_int t)
| Extr (t2, b) -> "t"^(string_of_int t2)^"."^(string_of_int b);;*)

let temp_string t =
  "t"^(string_of_int t);;

let temp_int t = t;;

let namer t = 
  try
    Hashtbl.find temp_names t
  with
    Not_found -> string_of_int (temp_int t)
;;

