(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

type action =
    PP | Interpret | AST | IC | ASS | SPIM | Liveness | Compile | INT

let the_action = ref Compile;;
let action() = !the_action;;
let do_action s = !the_action = s;;
let set a = Arg.Unit (fun () -> the_action := a);;

let all_files = ref [];;
let files() = !all_files;;
let talloc = ref false;;

let _ =
  Misc.option "-liveness" (set Liveness)
    "Montre le rsultat de liveness";
  Misc.option "-ass" (set ASS)
    "Montre le code spim avant allocation de registres (sans son prlude)";
  Misc.option "-ast" (set AST)
    "Montre l'arbre de syntaxte asbtraite";
  Misc.option "-pp" (set PP)
    "Imprime joliement le programme source";
  Misc.option "-ic" (set IC)
    "interprte le code intermediaire";
  Misc.option "-i" (set Interpret)
    "Interprte le programme";
  Misc.option "-spim" (set SPIM)
    "Produit le code spim (option par defaut)";
  Misc.option "-int" (set INT)
    "Produit le code intermdiaire";

  Arg.parse (Misc.options()) (fun s -> all_files := s :: !all_files) 

"Usage:

Prend le nom du fichier source en argument, et deux types d'options.
 * les options capitalises tracent une partie de l'excution dans stderr.
 * les autres indiquent le comportement du programme, et contrle la sortie
dans stdin, comme suit:
"
;;
(*
let print p =
  let show f =
    let g (_,_,c) = f c in
    g p.Spim.main; List.iter g (p.Spim.procedures) in
  if do_action SPIM || do_action ASS then
    show (List.iter (fun i -> Ass.nformat Spim.namer i; print_newline()));
  if do_action Liveness then
    show (fun c -> Liveness.print stdout Spim.namer (Liveness.flow c));
  p;;

let ( ** ) f g x = f (g x)
let spim = print ** Spim.program ** Canon.program ** Trans.program
let alloc = if !talloc then Talloc.program else Alloc.program
let compile = alloc ** spim

let print_code =
  List.iter (fun i -> Ass.nformat Spim.namer i; print_newline());;
*)
exception Error of string;;
let error s = raise (Error s);;

let parse_program error_prefix parseur lexer lexbuf =
  parseur lexer lexbuf
;;

let main() =
  try
    let chan, error_prefix =
      match files() with
      | s:: _ ->
          open_in s,
          String.concat "" [ "File \""; s; "\", line 1, characters ";]
      | [] ->
          stdin, "Characters " in
    let lexbuf = Lexing.from_channel chan in
    try
      Error.mylexbuf := lexbuf;
      let p = parse_program error_prefix Parser.main Lexer.token lexbuf in
        match action() with
        | PP -> Print.program p
	| AST ->  Ast.program p
        | INT -> ignore (Trans.trans_program p)
(*	| Interpret -> Interpret.eval p
	| IC -> Simul.program (Trans.program p)
	| ASS -> ignore (spim p)
	| _ -> print_code (compile p)*)
    with
      Error s -> Parsing.clear_parser(); prerr_string s; prerr_newline();
      exit 1
  with
  | Lexer.Eof -> () 
  | Sys_error s ->
      Printf.fprintf stderr "Erreur: %s" s; prerr_newline(); exit 1;
      ;;

Printexc.print main ();;

