(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(* Pick input file *)

(*let error_prefix, chan = 
  if (Array.length Sys.argv) = 1 then "", stdin
  else Sys.argv.(1), open_in Sys.argv.(1);;*)

let chan = Args.inchan 

(* Lex and parse *)

let lexbuf = Lexing.from_channel chan;;
Error.mylexbuf := lexbuf;;
let p = Parser.main Lexer.token lexbuf;;
    
if Args.verbose > 0 then begin
  print_string "\nPrinting AST.*************************************************\n";
  Ast.program p;
  print_string "\nEnd of AST.*************************************************\n\n";
  flush stdout;
end;;

(* Intermediate code generation *)

let intermed = Trans.trans_program p;;

if Args.verbose > 0 then begin
  print_string "\nPrinting intermediate.*************************************************\n";
  Code.output_list stdout intermed.Code.main.Code.code;
  print_newline ();
  List.iter (fun f -> Code.output_list stdout f.Code.code; print_newline ()) intermed.Code.funcs;
  print_string "\nEnd of intermediate.*************************************************\n\n";
  flush stdout;
end;;

(* Call extraction *)

Extractcalls.extractcalls_program intermed;;

(* Code generation and flow graph generation *)

let asm = Backend.emit_program intermed;;
  
if Args.verbose > 0 then begin
  print_string "\nPrinting assembly.*************************************************\n";
  Codegen.print_program_ass asm;
  print_string "\nEnd of assembly.*************************************************\n\n";
  print_string "\nPrinting flow graph.*************************************************\n";
  Liveness.print_program_flow asm;
  print_string "\nEnd of flow graph.*************************************************\n\n";
end;;
      
(* Register allocation *)

let fasm = Alloc.alloc Codegen.fast_registers Codegen.slow_registers Codegen.registers
    (List.fold_right (fun h l -> h.Ass.code @ l) asm [])
    (List.fold_right (fun h l -> h.Ass.flow @ l) asm []);;

if Args.verbose > 0 then begin
  print_string "\nPrinting colored asm.*************************************************\n";
  List.iter (fun i -> Printf.printf "%s\n" (Fasm.instr_string i)) fasm;
  print_string "\nEnd of colored asm.*************************************************\n\n";
end;;

(* Flow graph generation *)

let flowgraph = Traceprep.gengraph fasm;;
Traceprep.fill_maxsize flowgraph;;

if Args.verbose > 0 then begin
  print_string "\nPrinting flow graph.*************************************************\n";
  Traceprep.output_graph stdout flowgraph;
  print_string "\nEnd of flow graph.*************************************************\n\n";
end;;

(*Bank.simplify_banks flowgraph;;*)

Traceprep.triv_bank_simplify;;

let outchan = Args.outchan;;

Page.create_tree flowgraph outchan;;
close_out outchan;;
