%{
(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Pp;;
open Error;;
(*let globals = Hashtbl.create 100;;

let globaldef n v =
  if Hashtbl.mem globals n then parse_error ("name redefined :"(* ^ (Lexer.symbol_of_int n)*) )
  else Hashtbl.add globals n v;;
*)
let processvaritem typ h=
  match h.varexpression with
    NoneExpr -> (fst h.varident, VarExpr ((h.varident,fst typ),snd typ))
    | Expression hvarexpr -> (fst h.varident ,Sequence ((fst h.varident, VarExpr ((h.varident,fst typ),snd typ))
       ::[(fst hvarexpr, IExpression ( 
         (fst hvarexpr,ESet (h.varident,hvarexpr))
       ))]))
;;

let rec processvars typ n=
  begin match n with
      h::t -> (processvaritem typ h)::(processvars typ t)
    | [] -> []
  end
;; 


%}
%token <Pp.pos*Big_int.big_int> INT
%token <Pp.pos*int> IDENT
%token <Pp.pos*string> STRING
%token <Pp.pos*bool> BOOL
%token <Pp.pos*int> UINTEGER
%token <Pp.pos*int> SINTEGER
%token <Pp.pos> PLUS MINUS TIMES DIV MOD
%token <Pp.pos> LESSGREATER LESSEQUAL GREATEREQUAL LESS GREATER EQUALEQUAL
%token <Pp.pos> COLON EQUAL COMMA SEMI 
%token <Pp.pos> LPAREN RPAREN LBRACKET RBRACKET 
%token <Pp.pos> IF ELSE WHILE DO
%token <Pp.pos> BYTE BOOLEAN  
%token <Pp.pos> LACCOLADE,RACCOLADE
%token <Pp.pos> FOR SWITCH CASE VOID
%token  TEOF
%token <Pp.pos> INC DEC PLUSEQUAL MINUSEQUAL TIMESEQUAL DIVEQUAL MODEQUAL
%token <Pp.pos> LOGICALOR OR LOGICALAND AND XOR NOT NEG
%token <Pp.pos> SHL SHR ROL ROR RO4
%token <Pp.pos> ANDEQUAL OREQUAL XOREQUAL QUESTIONMARK
%token <Pp.pos> CONTINUE BREAK DEFAULT RETURN DOT DOTDOT STATIC COLONCOLON
%token <Pp.pos> TERA,TERB,TERC,TOPTION


/* lowest precedence */
%right else2
%nonassoc dot2
%right ELSE QUESTIONMARK COLON 
%right EQUAL,PLUSEQUAL,MINUSEQUAL,TIMESEQUAL,DIVEQUAL,ANDEQUAL,OREQUAL,XOREQUAL,MODEQUAL
%left LOGICALOR,LOGICALAND,LOGICALXOR
%left OR,AND,XOR
%left LESSGREATER GREATEREQUAL LESSEQUAL LESS GREATER EQUALEQUAL
%left PLUS MINUS 
%left TIMES DIV MOD
%nonassoc DOT DOTDOT COLONCOLON
%nonassoc UMINUS
%nonassoc LBRACKET 
%nonassoc cast

/* highest precedence */

/* the entry point */
%start main             
%type <Pp.program> main
%type <Pp.pinstruction list> definitions
%type <Pp.varline> varitem
%type <Pp.varline list> varlist
%type <Pp.pinstruction list> variable
%type <Pp.pos*Pp.expression> expression
%type <Pp.pinstruction> instruction
%type <Pp.var_list> args
%type <Pp.var_expr> argument
%type <Pp.var_list> args1
%type <Pp.pinstruction list> bloc
%%

epsilon :   { };

main:
  definitions TEOF    { $1 } 
;

definitions:
  variable SEMI definitions         { $1@$3  }
| procedure definitions             { $1::$2 }
| epsilon                           { [] }
;

variable:
  type_expr varitem varlist          { processvars ($1,false) ($2::$3)  }
  | STATIC type_expr varitem varlist { processvars ($2,true) ($3::$4)  }
;

varitem:
  IDENT                             { {varident=$1; varexpression=NoneExpr } }
| IDENT EQUAL expression            { {varident=$1; varexpression=Expression ((fst $3),(snd $3)) } }
;

varlist:
  COMMA varitem varlist             { $2::$3 }
| epsilon                           { [] }
;

procedure:
  type_expr IDENT LPAREN args RPAREN instruction
    { (fst $2, ProcDefinition { name=$2;  arguments = $4; result = Some $1; body = $6 }) }  
  | VOID IDENT LPAREN args RPAREN instruction
    { (fst $2, ProcDefinition { name=$2;  arguments = $4; result = None; body = $6 }) }   
;

args:
  VOID                                { [] }
  | argument args1                    { $1::$2 } 
  | epsilon                           { [] }
;

args1:
  COMMA argument args1                 { $2::$3 }
  | epsilon                           { [] }
;

argument:
  type_expr IDENT                     { $2,$1 }
;

instruction:
    LACCOLADE bloc RACCOLADE          { ($1,Bloc $2) }
  | FOR LPAREN ilist SEMI nexpression SEMI ilist RPAREN instruction 
    /* { ($1,Bloc ( ($2,Sequence $3)::[($4,While ($5,($6,Sequence[$9;($6,Sequence $7)])))] ) ) } */
                                      { ($1,For ($3,$5,$7,$9)) } 
  | WHILE LPAREN expression RPAREN instruction
                                      { ($1,For ([],$3,[],$5)) }
  | DO instruction WHILE LPAREN expression RPAREN SEMI
                                      { ($1,For ([],($1,Bool true),[(fst $5,If($5,($1,NoneInstruction),($4,Break)))],$2)) }
  | IF LPAREN expression RPAREN instruction %prec else2
                                      { ($1,If ($3,$5,($4,NoneInstruction)))  } 
  | IF LPAREN expression RPAREN instruction ELSE instruction 
                                      { ($1,If ($3,$5,$7)) } 
  | expression  SEMI                  { (fst $1,IExpression $1) }
  | SEMI                              { ($1,NoneInstruction) }
  | variable SEMI                     { ($2,Sequence $1) } 
  | CONTINUE SEMI                     { ($1,Continue) }
  | BREAK SEMI                        { ($1,Break) }
  | SWITCH LPAREN expression RPAREN LACCOLADE switchbloc RACCOLADE
                                      { ($1,Switch ($3,$6)) }
  | RETURN vexpression SEMI           { ($1,Return $2) }
; 

switchbloc:
  CASE const COLON bloc switchbloc    { {case=$2;inst=($3,Sequence $4)}::$5 }
  | DEFAULT COLON bloc                { [{case=($1,Default);inst=($2,Sequence $3)}] }
  | epsilon                           { [] }   
;

const:
  INT                                 { (fst $1,CInt (snd $1)) }
  | BOOL                              { (fst $1,CBool (snd $1)) }
;

ilist:
  expression ilist1                   { (fst $1,IExpression $1)::$2 }
  | variable                          { [ ({line=0;col=0},Sequence $1) ] }
  | epsilon                           { [] }
;

ilist1:
  COMMA expression ilist1             { ($1,IExpression $2)::$3 }
  | COMMA variable                    { [ ($1,Sequence $2) ] }
  | epsilon                           { [] } 
;

bloc: 
  instruction bloc                    { $1::$2 }
  | epsilon                           { [] }
;

nexpression:
  expression                          { $1 }
  | epsilon                           { ({ line=0;col=0},Bool true) }
;

vexpression:
  expression                          { Expression $1 }
  | epsilon                           { NoneExpr  }
;

expression:
  LPAREN expression RPAREN            { $2 }
| INT                                 { (fst $1,Int (snd $1)) }
| BOOL                                { (fst $1,Bool (snd $1)) }
| IDENT                               { (fst $1,Get $1) }

| MINUS expression %prec UMINUS       { ($1,Un (Neg,$2)) }
| NEG expression %prec UMINUS         { ($1,Un (Not,$2)) }
| NOT expression %prec UMINUS         { ($1,Un (Not,$2)) }
| SHL expression %prec UMINUS         { ($1,Un (Shl,$2)) }
| SHR expression %prec UMINUS         { ($1,Un (Shr,$2)) }
| ROL expression %prec UMINUS         { ($1,Un (Rol,$2)) }
| ROR expression %prec UMINUS         { ($1,Un (Ror,$2)) }
| RO4 expression %prec UMINUS         { ($1,Un (Ro4,$2)) }

| expression PLUS expression          { ($2,Bin (Plus, $1,  $3)) }
| expression MINUS expression         { ($2,Bin (Minus, $1,  $3)) }
| expression TIMES expression         { ($2,Bin (Times, $1, $3)) }
| expression DIV expression           { ($2,Bin (Div,  $1,  $3)) }
| expression MOD expression           { ($2,Bin (Mod,  $1,  $3)) }

| expression LESS expression          { ($2,Bin (Lt, $1, $3)) }
| expression LESSEQUAL expression     { ($2,Bin (Le, $1, $3)) }
| expression GREATER expression       { ($2,Bin (Gt, $1, $3)) }
| expression GREATEREQUAL expression  { ($2,Bin (Ge, $1, $3)) }
| expression EQUALEQUAL expression    { ($2,Bin (Eq, $1, $3)) }
| expression LESSGREATER expression   { ($2,Bin (Ne, $1, $3)) }

| expression OR  expression           { ($2,Bin (Or, $1, $3)) }
| expression AND expression           { ($2,Bin (And, $1, $3)) }
| expression XOR expression           { ($2,Bin (Xor, $1, $3)) }
| expression COLONCOLON expression    { ($2,Bin (Concat , $1, $3)) }

| expression LOGICALOR  expression    { ($2,Bin (L_Or, $1, $3)) }
| expression LOGICALAND expression    { ($2,Bin (L_And, $1, $3)) }
| expression QUESTIONMARK expression COLON expression 
                                      { ($2,Ifte($1,$3,$5))  } 
| IDENT EQUAL expression              { ($2,ESet($1,$3)) }
| IDENT PLUSEQUAL expression          { ($2,ESet($1,($2,Bin (Plus ,  (fst $1,Get $1), $3)))) }
| IDENT MINUSEQUAL expression         { ($2,ESet($1,($2,Bin (Minus , (fst $1,Get $1), $3)))) }
| IDENT TIMESEQUAL expression         { ($2,ESet($1,($2,Bin (Times , (fst $1,Get $1), $3)))) }
| IDENT DIVEQUAL expression           { ($2,ESet($1,($2,Bin (Div ,   (fst $1,Get $1), $3)))) }
| IDENT MODEQUAL expression           { ($2,ESet($1,($2,Bin (Mod ,   (fst $1,Get $1), $3)))) }
| IDENT ANDEQUAL expression           { ($2,ESet($1,($2,Bin (And , (fst $1,Get $1), $3)))) }
| IDENT OREQUAL expression            { ($2,ESet($1,($2,Bin (Or ,  (fst $1,Get $1), $3)))) }
| IDENT XOREQUAL expression           { ($2,ESet($1,($2,Bin (Xor , (fst $1,Get $1), $3)))) }
| IDENT INC                           { ($2,Incr(0,1,$1) ) }
| IDENT DEC                           { ($2,Incr(0,-1,$1) ) }
| INC IDENT                           { ($1,Incr(1,1,$2) ) }
| DEC IDENT                           { ($1,Incr(1,-1,$2) ) }
| expression DOT INT %prec dot2       { ($2,Un (Extractbit (Big_int.int_of_big_int (snd $3)),$1)) }
| expression DOT INT DOTDOT INT       { ($2,Un (Extract (Big_int.int_of_big_int (snd $3),Big_int.int_of_big_int (snd $5)),$1)) } 
| LPAREN type_expr RPAREN expression %prec cast  { ($1,Un (Cast $2,$4)) }
| IDENT LPAREN fctarguments RPAREN    { (fst $1,Function_call ($1,$3)) } 

| TERA EQUAL expression               { ($1,Un (SAssign ERA,$3)) }
| TERB EQUAL expression               { ($1,Un (SAssign ERB,$3)) }
| TERC EQUAL expression               { ($1,Un (SAssign ERC,$3)) }
| TOPTION EQUAL expression            { ($1,Un (SAssign OPTION,$3)) }

;

fctarguments:
  expression  fctarguments1           { $1 :: $2 }
| epsilon                             { [] }
;
fctarguments1:
  COMMA expression fctarguments1      { $2 :: $3 }
| epsilon                             { [] }
;

type_expr:
  BYTE                                { Byte }
| UINTEGER                            { UInteger (snd $1) }
| SINTEGER                            { SInteger (snd $1) }
| BOOLEAN                             { Boolean }
;


