(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

type 'a t = { all : 'a elem list array; id : int }
and 'a elem = { info : 'a; mutable parent : 'a t }
let make n =
  let all = Array.create n [] in
  Array.mapi (fun i _ -> { all = all; id = i }) all
let create s e =
  let n = { info = e; parent = s } in
  s.all.(s.id) <- n :: s.all.(s.id);
  n
let eqsets s1 s2 = (s1.all == s2.all && s1.id = s2.id)
let eq u v = (u == v)
let parent e = e.parent
let info e = e.info
let clear s = s.all.(s.id) <- []
    
let belong e s = eqsets e.parent s

let move e s =
  if eqsets e.parent s then ()
  else (e.parent <- s; s.all.(s.id) <- e::s.all.(s.id))

let pick s =
  let rec next = function
    | h::t as l ->
        if eqsets h.parent s
        then (s.all.(s.id) <- l; h)
        else next t
    | [] -> raise Not_found 
  in next s.all.(s.id)
;;

let empty s = try let _ = pick s in false with Not_found -> true

let list s =
  let rec next = function
    | h::t as l ->
        if eqsets h.parent s
        then h :: next t
        else next t
    | [] -> [] in
  let l = next s.all.(s.id) in
  s.all.(s.id) <- l;
  l
;;

let pick_lowest cost s =
  let m = ref (pick s) in
  let rec find_and_clean q = function
    | h::t ->
        if eqsets h.parent s
        then
          let q' = cost h in
          let q'' = if q' < q then (m := h; q') else q in
          h :: find_and_clean q'' t
        else
          find_and_clean q t
    | [] -> [] in
  let l = find_and_clean (cost !m) s.all.(s.id) in
  s.all.(s.id) <- l;
  !m
    ;;

let sort f s = s.all.(s.id) <- Sort.list f (list s)
    ;;

