(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(* Syntaxe abstraite des expression de type *)
type pos = {
    line : int;
    col : int;
  };;
type type_expr = Byte | SInteger of int | UInteger of int | Boolean | TConst of Big_int.big_int | Void;;
type ident = (pos*int);;
type var_expr = (ident * type_expr);; 
type var_list = var_expr list;;

let pos_string ps = 
  Format.sprintf "at line %d, col %d" ps.line ps.col;
;;

let posdummy () = {line =0;col=0};

type program = pinstruction list
  
and procdefinition = {
    name : ident;
    arguments : var_list; 
    result : type_expr option;
    (* arguments  et type du rsultat *)
    body : pinstruction;
    (* corps de la fonction *)
  }

and expression =
  | Int of Big_int.big_int | Bool of bool
    (* constantes *)
  | Bin of binop * pexpression * pexpression  (**)
    (* expressions arithtiques *)
  | Un of unop * pexpression
  | Get of ident
    (* accs  une variable *)
  | Function_call of ident * pexpression list 
    (* appel de fonction *)
  | ESet of ident * pexpression
    (* Affectation d'une variable  l'interieur d'une expression *)
  | Incr of int*int*ident
  | Ifte of pexpression * pexpression * pexpression 

and binop =
  | Plus | Minus | Times | Div | Mod
    (* arithmtique *)
  | Lt | Le | Gt | Ge | Eq | Ne
    (* comparaisons *) 
  | Xor | Or | And
    (* Arithmetic&logic operators *)
  | L_Or | L_And | Concat
and unop = Neg | Not | Shr | Shl | Rol | Ror | Ro4 | Cast of type_expr
  | Extract of int * int
  | Extractbit of int 
  | SAssign of specialregister 
and specialregister =  ERA | ERB | ERC | OPTION
  
and instruction = 
  | IExpression of pexpression
  | Bloc of pinstruction list
  | Sequence of pinstruction list
  | For of (pinstruction list) * pexpression * (pinstruction list) * pinstruction
  | Switch of pexpression * (switchstatement list) (**)
  | If of pexpression * pinstruction * pinstruction
  | VarExpr of var_expr*bool
  | ProcDefinition of procdefinition
  | Continue
  | Break
  | Return of vexpression
  | NoneInstruction
and switchstatement = {
    case : pos*const; 
    inst : pinstruction;  
  }
and const=
    CInt of Big_int.big_int | CBool of bool | Default
and varline = {
    varident : ident;
    varexpression : vexpression;
  }
and vexpression = 
    Expression of pexpression
  | NoneExpr 
and pexpression = pos*expression
and pinstruction = pos*instruction
 

