/*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

byte readRC()
{
  return RC;
}

void writeRB(byte val)
{
  RB = val;
}

byte sumofint(byte val)
{
  byte ret;
 
  for (ret = 0; val != 0; ret += val, val--);

  return ret;
}

byte fibb(byte val)
{
  byte a = 1, b = 1, c;
  
  for (; val > 0; val--)
  {
    c = a + b;
    a = b;
    b = c;
  }

  return a;
}

byte calc(byte val)
{
  return fibb(val); 
}

void pwm(byte val)
{
  byte c = 0xFF;

  for (RB = 0; c > val; c--);
  for (RB = 0xFF; c > 0; c--);
}

uint16 multu8u8(uint8 a, uint8 b)
{
  byte n;
  uint16 result = 0;
  uint16 bext = b;
  for (n = 0; n < 4; n++)
  {
    if (a.0)
    {
      result = result + bext;
      //      break;
    }
  
    a = shr a;
    bext = shl bext;
  }

  return result;
}

uint8 divu8u8(uint8 a, uint8 b)
{
  uint16 sub = 0;
  byte n;
  byte rslt = 0;

  sub = (b::0);

  for (n = 0; n < 8; n++)
  {
    rslt = shl rslt;
    sub = shr sub;

    if (sub <= a)
    {
      a -= (byte) sub;
      rslt++;
    }
  }    
  
  return rslt;
}

uint8 modu8u8(uint8 a, uint8 b)
{
  uint16 sub = 0;
  byte n;

  sub = (b::0);

  for (n = 0; n < 8; n++)
  {
    sub = shr sub;

    if (sub <= a)
      a -= (byte) sub;
  }    
  
  return a;
}

uint8 findfactor(uint8 a)
{
  uint8 n;

  if (((byte) a.0) == 0) return 2;

  for (n = 3; n < 16; n+=2)
  {
    if (modu8u8(a, n) == 0) 
      return n;
  }

  return 1;
}

void main()
{
  M = 0x0F;
  !RB = 0;
  !RA = 0;

  while (1)
  {
    uint8 p = RC;
    uint8 f = findfactor(p);
    RA = f;
    RB = divu8u8(p, f);
  }
}
