(*
  SXC: Compiler for a C-like language for the SX (Ubicom) microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>

  This code was almost entirely rewritten form an initial skeleton by Didier 
  Rmi, some of the original code probably remains.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Code;;
open Format;;

let transerrors=ref 0;;

type env = {
  vars : (int,int) Hashtbl.t;
};;


type procedure = {
  mutable pid : int; (*identifiant unique de la procedure*)
  pname : int; (*index dans la hashtable du lexer*)
  procdef : Pp.procdefinition;
  mutable funct : func;
};;


let variabletable = Hashtbl.create 100;; (*association id->variable*)
let proctable = Hashtbl.create 100;; (*association id->procedure*)
let procnametable = Hashtbl.create 100;; (*association id_nom->id*)
let currentenv = ref [{vars=(Hashtbl.create 100)}];; (*association  id_nom ->id*)
let lastid = ref (-1);;
let lastprocid = ref(-1);;
let funclist = ref [];;
let usefunct = ref [];;
let staticlist = ref [];;
let procresult = ref None;;
let lastbreak = ref None;;
let lastcontinue = ref None;;

(*-----------------------------------------------------------------------------------------------*)
let big_int_to_const bigint =
      let nb=ref 0 in
      let bignb=ref bigint in          
      let signed=ref false in
      let lst=ref [] in
	begin
        signed:=(Big_int.lt_big_int !bignb Big_int.zero_big_int ); 
        if (!signed) then bignb:=Big_int.sub_big_int (Big_int.big_int_of_int (-1)) !bignb ;
        while (not (Big_int.eq_big_int Big_int.zero_big_int !bignb ) ) do
          lst:=!lst@[(function a -> if !signed then 255-a else a)  (Big_int.int_of_big_int (Big_int.mod_big_int !bignb (Big_int.big_int_of_int 256)))];
          bignb:=Big_int.div_big_int !bignb (Big_int.big_int_of_int 256); 
          incr nb;
        done;
        if (!nb=0) then ( lst:=[if !signed then 255 else 0];nb:=1; );  
        if (Args.verbose>=2) then (
          printf "Int declaration %s : [" (Big_int.string_of_big_int bigint);
          List.iter (function a->(printf "%d " a;)) !lst; 
          printf "]\n";
        );
        Const !lst;
	end

let big_int_in_range bigint big1 big2 =(Big_int.le_big_int big1 bigint)&&(Big_int.ge_big_int big2 bigint);;

let tobig i=Big_int.big_int_of_int i;;

let listextract i j l = 
  let rec extract n = function
      l when n = j + 1 -> []
    | [] -> invalid_arg "listextract"
    | h::t -> let r = extract (n + 1) t in 
      if n >= i then h::r else r
  in
  extract 0 l;;

let repet item n = Array.to_list (Array.make n item);;

let expand_const vl mycst nn=
  match mycst with
  Const cst1-> (  
    let n1=List.length cst1 in 
    let n=nn/8 in
    if (n1=n) then (Const cst1) else
    if (n1>n) then (Const (listextract 0 (n-1) cst1)) else  (  
      (Const ( cst1@(repet vl (n-n1)) ) )
    )
  )
    | _ -> (Const [0]);;  (*ne doit jamais arriver*)

let truncated_bigint bigint n= (
  (Big_int.mod_big_int bigint (Big_int.power_int_positive_int 2 n))
)

let extractconst csta =
  match csta with
    Const a -> a;
    |_ ->  [0];
;;


let get_bigint_mini_type bigint = (
  if (Big_int.ge_big_int bigint (tobig 0)) then
    (  Pp.UInteger (8*(List.length (extractconst (big_int_to_const bigint))))      
    ) else (
       Pp.SInteger (8*(List.length (extractconst (big_int_to_const bigint))))
    )
)
(*-----------------------------------------------------------------------------------------------*)

let var_can_define s =  (* dtermine si la variable est dans l'environnement courant
 si oui renvoie l'id global, sinon -1 *)
  try Hashtbl.find (List.hd (!currentenv)).vars s 
  with Not_found -> -1
;;

let rec var_exists env s = (*determine si la variable est dans un des environnements successifs si oui renvoie id global, sinon -1 *)
  match env with 
    [] -> -1
  | h::t -> begin
    try Hashtbl.find h.vars s 
    with Not_found -> var_exists t s 
  end
;;

let push_env () =
  currentenv:={vars=Hashtbl.create 100}::(!currentenv);  
;;

let pop_env () = 
  match !currentenv with
    [] -> ();
    | h::t-> currentenv:=t;     
;;

let rec sizetemps vtyp= (*renvoie le nombre de bytes ncessaires pour le type de variable donn*)
  match vtyp with
    Pp.Byte -> 1
    | Pp.Boolean -> 1
    | Pp.UInteger n -> (n/8) 
    | Pp.SInteger n -> (n/8)  
    | Pp.Void -> 0
    | Pp.TConst bigint -> sizetemps (get_bigint_mini_type bigint)
;;

let alloctemps vtyp b =
  let n=sizetemps vtyp in
  let list=ref [] in
  begin
    for i = 1 to n do 
      let item=(Gen.new_temp()) in
      begin
        list:=item::(!list);
        if b then staticlist:=item::(!staticlist); 
      end;
    done;
    !list;
  end
;;

let createvar v vpos=  (*cre une variable v dans l'environnement courant si possible,sinon error*)
  let id=var_can_define v.vname in
  if (id=(-1)) then begin
    incr lastid; 
    v.vid<-(!lastid);
    (match (v.typ) with
      Pp.UInteger n -> if ((n<8)||((n mod 8)!=0)) then ( incr transerrors;Printf.fprintf stderr "Error : Invalid type %s\n" (Pp.pos_string vpos);v.typ<-Pp.UInteger 8);
      | Pp.SInteger n -> if ((n<8)||((n mod 8)!=0)) then ( incr transerrors;Printf.fprintf stderr "Error : Invalid type %s\n" (Pp.pos_string vpos);v.typ<-Pp.UInteger 8);
      | _ -> ();
    );
    if not (v.vname=(-2)) then Hashtbl.add (List.hd (!currentenv)).vars v.vname (!lastid);
    Hashtbl.add variabletable (!lastid) v;
    if (Args.verbose>=2) then ( 
      printf "Declare var %s with vid %d %s\n  " (Lexer.symbol_of_int v.vname) (!lastid) (Pp.pos_string vpos);
      List.iter (function a -> printf "%d " a ) v.temps;
      printf "\n";
    )
  end else begin
    incr transerrors;Printf.fprintf stderr "Error : Redeclared variable %s %s\n" (Lexer.symbol_of_int v.vname) (Pp.pos_string vpos); 
  end; 
;;

let createtemp vtyp= (*cre une variable temporaire*)
  let v={vid=0;vname=(-2);typ=vtyp;static=false;temps=alloctemps vtyp false} in
  begin
    createvar v (Pp.posdummy());
    v;
  end
;;

let getvar ident = (*renvoie la variable de nom ident.name *)
  let id=var_exists (!currentenv) (snd ident) in (
    if (id=(-1)) then begin
      printf "Error : Unknown variable \"%s\" %s\n" (Lexer.symbol_of_int (snd ident)) (Pp.pos_string (fst ident));
      incr transerrors;
      Printf.fprintf stderr "Error : Unknown variable \"%s\" %s\n" (Lexer.symbol_of_int (snd ident)) (Pp.pos_string (fst ident));
      { vid=0;vname=(snd ident);typ=Pp.Byte;static=false;temps=alloctemps Pp.Byte false};
    end else begin
      if (Args.verbose>=2) then printf "use var %s with vid %d %s\n" (Lexer.symbol_of_int (snd ident)) id (Pp.pos_string (fst ident));
      Hashtbl.find variabletable id;
    end;   
    )
;;
(*------------------------------------------------------------------------------------------------*)
let proc_can_define s =  (* dtermine si la procedure existe si oui renvoie l'id global, sinon -1 *)
  try Hashtbl.find procnametable s
  with Not_found -> -1
;;

let createproc p (ppos:Pp.pos)=  (*cre une procedure p si possible,sinon error*)
  let id=proc_can_define p.pname in
  if (id=(-1)) then begin
    incr lastprocid; 
    p.pid<-(!lastprocid);
    Hashtbl.add procnametable p.pname (!lastprocid);
    Hashtbl.add proctable (!lastprocid) p;
    if (Args.verbose>=2) then printf "Declare proc %s with pid %d %s\n" (Lexer.symbol_of_int p.pname) (!lastprocid) (Pp.pos_string ppos);
  end else begin
    incr transerrors;Printf.fprintf stderr "Error : Redeclared procedure %s %s\n" (Lexer.symbol_of_int p.pname) (Pp.pos_string ppos); 
  end;
;;

let nullfunc ()=
  { code=[];name={Gen.id=(-2);Gen.name="unknownfunct"};args=[];result=[];resulttype=Pp.Byte;statics=[];uses=[];defs=[];usefuncs=[];translating=false;translated=false;depth=0;}
;;

let currentfunc = ref (nullfunc ());;

let getproc ident = (*renvoie la procedure de nom ident.name *)
  let id=proc_can_define (snd ident) in
    if (id=(-1)) then begin
      incr transerrors;
      Printf.fprintf stderr "Error : Unknown procedure %s %s\n" (Lexer.symbol_of_int (snd ident)) (Pp.pos_string (fst ident));
      { pid=0;pname=0;procdef={Pp.name=ident;Pp.arguments=[];Pp.result=None;Pp.body=(fst ident,Pp.NoneInstruction)};funct=nullfunc () };
    end else begin
      if (Args.verbose>=2) then printf "use proc %s with pid %d %s\n" (Lexer.symbol_of_int (snd ident)) id (Pp.pos_string (fst ident));
      let proc=Hashtbl.find proctable id in
      begin 
        usefunct:=proc.funct::(List.filter (function a->(a!=proc.funct)) (!usefunct)); 
        proc; 
      end; 
    end;   
;;
(*-----------------------------------------------------------------------------------------------------------------------------*)

let get_compatible_type typ1 typ2 epos = (* prend deux Pp.type_expr et une pos d'erreur renvoie un Pp.type_expr compatible*) 
  let ltyp1=(match typ1 with 
	       | Pp.TConst bigint -> get_bigint_mini_type bigint
	       | _                -> typ1) in
  let ltyp2=(match typ2 with
	       | Pp.TConst bigint -> get_bigint_mini_type bigint
	       | _                -> typ2) in           
  match (ltyp1,ltyp2) with
    | (Pp.Boolean,Pp.Boolean)    -> Pp.Boolean;
    | (Pp.Byte,Pp.Boolean)       -> Pp.Byte;
    | (Pp.UInteger n,Pp.Boolean) -> Pp.UInteger n;
    | (Pp.SInteger n,Pp.Boolean) -> Pp.SInteger n;
     (**************)
    | (Pp.Boolean,Pp.Byte)       -> Pp.Byte;
    | (Pp.Byte,Pp.Byte)          -> Pp.Byte;
    | (Pp.UInteger n,Pp.Byte)    -> Pp.UInteger n;
    | (Pp.SInteger n,Pp.Byte)    -> Pp.SInteger n;
    (**************)
    | (Pp.Boolean,Pp.UInteger n) -> Pp.UInteger n;
    | (Pp.Byte,Pp.UInteger n)    -> Pp.UInteger n;
    | (Pp.UInteger n1,Pp.UInteger n2) -> Pp.UInteger (if n1>n2 then n1 else n2)
    | (Pp.SInteger n1,Pp.UInteger n2) -> Pp.SInteger (if n1>n2 then n1 else n2)
    (**************)
    | (Pp.Boolean,Pp.SInteger n) -> Pp.SInteger n;
    | (Pp.Byte,Pp.SInteger n)    -> Pp.SInteger n;
    | (Pp.UInteger n1,Pp.SInteger n2) -> Pp.SInteger (if n1>n2 then n1 else n2) 
    | (Pp.SInteger n1,Pp.SInteger n2) -> Pp.SInteger (if n1>n2 then n1 else n2)
    (**************)
    | (Pp.Void,t) -> t;
    | (t,Pp.Void) -> t;
    | (_,_) -> incr transerrors;Printf.fprintf stderr "get_compatible_type error\n";Pp.Byte;
;;  

let rec nothing_of_type t = (setcast (Const [0],Pp.Byte) t (Pp.posdummy ()) false) 

and setcast (exp1,typ1) typ2 epos err=  (*prend un couple Code.exp*Pp.type_expr + un bool  et un Pp.type_expr puis renvoie un Code.exp*Pp.type_expr*)
  match (typ1,typ2) with
    (Pp.Boolean,Pp.Boolean) -> (exp1,typ2);
    | (Pp.TConst bigint,Pp.Boolean) -> (
        if (Big_int.eq_big_int (tobig 0) bigint) then (Const [0],Pp.TConst (tobig 0)) else (Const [1],Pp.TConst (tobig 1)) 
      );
    | (_,Pp.Boolean) -> (Code.Un (NotZero,exp1),typ2);
     (**************)
    | (Pp.Byte,Pp.Byte) -> (exp1,typ2);
    | (Pp.Boolean,Pp.Byte) -> (exp1,typ2);
    | (Pp.TConst bigint,Pp.Byte) -> (
        if (err)&&(not (big_int_in_range bigint (tobig 0) (tobig 255))) then Printf.fprintf stderr "Warning : Constant truncated %s\n" (Pp.pos_string epos);
        let i=List.hd (extractconst exp1) in
          ( Const [i] , Pp.TConst (tobig i) )
      );
    | (Pp.UInteger n,Pp.Byte) -> begin
        if ((n>8)&&(err)) then Printf.fprintf stderr "Warning : Implicit cast reduces datawidth %s\n" (Pp.pos_string epos);
        (Un (Extract (0,0),exp1),typ2);    
      end;
    | (Pp.SInteger n,Pp.Byte) -> begin
        if (err) then Printf.fprintf stderr "Warning : implicit cast from signed to unsigned %s\n" (Pp.pos_string epos);
        (Un (Extract (0,0), exp1),typ2);    
      end;
    (**************)
    | (Pp.Boolean,Pp.UInteger n) -> (Un ((UExtend ((n/8)-1)),exp1 ),typ2);
    | (Pp.Byte,Pp.UInteger n) -> if (n=8) then (exp1,typ2) else (Un ((UExtend ((n/8)-1)),exp1 ),typ2);
    | (Pp.TConst bigint,Pp.UInteger n) -> (
        if not (big_int_in_range bigint (tobig 0) (Big_int.add_int_big_int (-1) (Big_int.power_int_positive_int 2 n))) then Printf.fprintf stderr "Warning : Constant truncated %s\n" (Pp.pos_string epos);
        ( expand_const 0 exp1 n , Pp.TConst (truncated_bigint bigint n) )
      );
    | (Pp.SInteger n1,Pp.UInteger n2) -> begin
        if (err) then Printf.fprintf stderr "Warning : implicit cast from signed to unsigned %s\n" (Pp.pos_string epos);
        if (n2>n1) then (Un ((SExtend ((n2-n1)/8)),exp1 ),typ2)
        else begin
          if (n1=n2) then (exp1,typ2)
          else begin
            if (err) then Printf.fprintf stderr "Warning : implicit cast reduces datawidth %s\n" (Pp.pos_string epos);
            (Un (Extract (0,(n2/8)-1),exp1),typ2);      
          end;
        end; 
      end;
    | (Pp.UInteger n1,Pp.UInteger n2) -> begin
        if (n2>n1) then (Un (UExtend ((n2-n1)/8),exp1 ),typ2)
        else begin
          if (n1=n2) then (exp1,typ2)
          else begin
            if (err) then Printf.fprintf stderr "2Warning : implicit cast reduces datawidth %s\n" (Pp.pos_string epos);
            (Un (Extract (0,(n2/8)-1), exp1),typ2);      
          end;
        end; 
      end; 
    (**************)
    | (Pp.Boolean,Pp.SInteger n) -> (Un (UExtend ((n/8)-1),exp1 ),typ2);
    | (Pp.Byte,Pp.SInteger n) -> (
        if err then Printf.fprintf stderr "Warning : implicit cast from unsigned to signed %s \n" (Pp.pos_string epos);
        if (n=8) then (exp1,typ2) else (Un (UExtend ((n/8)-1),exp1 ),typ2);
      )
    | (Pp.TConst bigint,Pp.SInteger n) -> (
        if not (big_int_in_range bigint (Big_int.minus_big_int (Big_int.power_int_positive_int 2 (n-1))) (Big_int.add_int_big_int (-1) (Big_int.power_int_positive_int 2 (n-1)) )) then Printf.fprintf stderr "Warning : Constant truncated %s\n" (Pp.pos_string epos);
          ( if (Big_int.ge_big_int  bigint Big_int.zero_big_int) then 
                 (expand_const 0 exp1 n,Pp.TConst (truncated_bigint bigint n)) 
             else 
                 (expand_const 255 exp1 n,Pp.TConst (truncated_bigint bigint n)) 
          )
      );
    | (Pp.UInteger n1,Pp.SInteger n2) -> begin
        if (n1>=n2)&&(err) then Printf.fprintf stderr "Warning : implicit cast from unsigned to signed %s\n" (Pp.pos_string epos);
        if (n2>n1) then (Un (UExtend ((n2-n1)/8),exp1 ),typ2)
        else begin
          if (n1=n2) then (exp1,typ2)
          else begin
            if (err) then Printf.fprintf stderr "Warning : implicit cast reduces datawidth %s\n" (Pp.pos_string epos);
            (Un (Extract (0,(n2/8)-1),exp1),typ2);      
          end;
        end; 
      end;
    | (Pp.SInteger n1,Pp.SInteger n2) -> begin
        if (n2>n1) then (Un (SExtend ((n2-n1)/8),exp1 ),typ2)
        else begin
          if (n1=n2) then (exp1,typ2)
          else begin
            if (err) then Printf.fprintf stderr "Warning : implicit cast reduces datawidth %s\n" (Pp.pos_string epos);
            (Un (Extract (0,(n2/8)-1),exp1),typ2);      
          end;
        end; 
      end; 
    | (Pp.Void,t) -> incr transerrors;Printf.fprintf stderr "Error : Incompatible types void and %s %s\n" (Ast.type_expr t) (Pp.pos_string epos);nothing_of_type t;
    | (t,Pp.Void) -> incr transerrors;Printf.fprintf stderr "Error : Incompatible types void and %s %s\n" (Ast.type_expr t) (Pp.pos_string epos);nothing_of_type t;
    | (_,(_ as t)) -> incr transerrors;Printf.fprintf stderr "setcast error\n";nothing_of_type t;
;;

let rec processbloc il= (*renvoie Code.stm list*)
  push_env ();
  let result=List.fold_left add_p_instruction [] il in
  begin
    pop_env ();
    result;
  end

and oper_same_type op exp1 exp2 =
  let expr1=p_expression exp1 in
  let expr2=p_expression exp2 in 
  let typ=get_compatible_type (snd expr1) (snd expr2) (fst exp1) in 
  let expr2b=setcast (expr1) typ (fst exp1) true in
  let expr3b=setcast (expr2) typ (fst exp2) true in 
    (Bin (op (),fst expr2b,fst expr3b),typ) 

and oper_bool op exp1 exp2 =
  let expr1=p_expression exp1 in
  let expr2=p_expression exp2 in 
  let typ=Pp.Boolean in 
  let expr2b=setcast (expr1) typ (fst exp1) true in
  let expr3b=setcast (expr2) typ (fst exp2) true in 
    (Bin (op (),fst expr2b,fst expr3b),typ) 

and oper_sign_homogen opsigned opunsigned exp1 exp2 =
  let expr1=p_expression exp1 in
  let expr2=p_expression exp2 in 
  let typ=get_compatible_type (snd expr1) (snd expr2) (fst exp1) in 
  let expr2b=setcast (expr1) typ (fst exp1) true in
  let expr3b=setcast (expr2) typ (fst exp2) true in 
  ( 
    match typ with 
      | Pp.SInteger n -> (Bin (opsigned (),fst expr2b,fst expr3b),typ) 
      | _ -> (Bin (opunsigned (),fst expr2b,fst expr3b),typ) 
    )
and declarevar v b= (* var_expr * boolean -> () *)
  let temps = alloctemps (snd v) b in
  begin
    ignore (List.fold_left (fun n t -> 
      Gen.name_temp t ((Lexer.symbol_of_int (snd (fst v)))^"["^(string_of_int n)^"]");
      (n + 8)) 0 temps);
    let var={ vid=0;vname=snd (fst v);typ=snd v;static=b;temps=temps}  in (
      createvar var (fst (fst v));
      var;
    )
  end
and addargument lst p=
let (id,atyp)=p in
  let var=declarevar p false (*{ vid=0;vname=(-1);typ=atyp;static=false;temps=alloctemps atyp false}*) in
   lst@[var.temps] 
and declareproc (p:Pp.procdefinition)=
  usefunct:=[];
  staticlist:=[];
  push_env ();
 let result= (
  let arglist = List.fold_left addargument [] p.Pp.arguments in
  let resultlist=
  (match p.Pp.result with
    None -> procresult := None; [];
    | Some a -> (
         (procresult:=Some (createtemp a);((function Some a -> a) !procresult).temps;)
      )
  ) in
  let firstlabel = Gen.named_label (Lexer.symbol_of_int (snd p.Pp.name)) in
  let myfunc= (currentfunc:={ code = []; (*rempli plus loin*) 
                name = firstlabel; 
                args = arglist;
                result = resultlist; 
                resulttype = (if resultlist=[] then Pp.Byte else ((function somea -> match somea with Some a->a|None -> Printf.fprintf stderr "func.resulttype error";Pp.Byte) p.Pp.result) );
                statics=[]; (*rempli plus loin*)
                uses=[];
                defs=[];
                usefuncs = []; (*rempli plus loin*)
                translating =false;
                translated=false;
		depth=0;	    
	      };!currentfunc) in                     
  let proc = { pid=0;pname=snd p.Pp.name;procdef=p;funct=myfunc } in
  begin
    createproc proc (fst p.Pp.name);
    let proccode= (Label firstlabel)::(p_instruction p.Pp.body) in               
    begin
      funclist:=myfunc::(!funclist); (* ordre peu important ici *)
      myfunc.code <- proccode@[Return (!currentfunc,Temp resultlist)];
      myfunc.usefuncs <- !usefunct;
      myfunc.statics <- !staticlist;
    end;
  end;
 ) in begin
   pop_env();
   result;
 end
and add_p_instruction lst p =  (* (Code.stm list)*pinstruction -> (Code.stm list) *)
  lst@(p_instruction p);
and p_instruction p =  (* renvoie  Code.stm list *)
  match (snd p) with 
    | Pp.IExpression expr -> [Exp (fst (p_expression expr))];
    | Pp.Bloc il -> processbloc il;
    | Pp.Sequence il -> List.fold_left add_p_instruction [] il;
    | Pp.Switch (a,b) -> incr transerrors;Printf.fprintf stderr "Error : feature Switch not implemented %s\n" (Pp.pos_string (fst p));[];
    | Pp.For (instl1,forexpr,instl2,instloop) -> (
      push_env ();
      let savelastbreak=(!lastbreak) in
      let savelastcontinue=(!lastcontinue) in
      let labelloop=Gen.new_label () in
      let labelnext=Gen.new_label () in
      let labellooptest=Gen.new_label () in
      let labelend=Gen.new_label () in 
      let instrinit=(List.fold_left add_p_instruction [] instl1) in	
      let condition=fst (setcast (p_expression forexpr) Pp.Boolean (fst forexpr) true) in
      let result=(
        lastbreak:=Some labelend;
        lastcontinue:=Some labelnext; 
        instrinit@
        [Jump labellooptest;
         Label labelloop]@
        (p_instruction instloop)@
        [Jump labelnext;
         Label labelnext]@
        (List.fold_left add_p_instruction [] instl2)@
        [Jump labellooptest;
         Label labellooptest;
         Cjump (condition,true,labelloop,labelend );
         Label labelend]
      ) in (
        lastbreak:=savelastbreak;
        lastcontinue:=savelastcontinue;  
	pop_env ();
        result; 
	)
      );
    | Pp.If (ifexpr,inst1,inst2) -> 
      let label1=Gen.new_label () in 
      let label2=Gen.new_label () in 
      let endlabel=Gen.new_label () in 
      [Cjump (fst (setcast (p_expression ifexpr) Pp.Boolean (fst ifexpr) true),true,label1,label2 );
       Label label2;]@
       (p_instruction inst2)
       @[ Jump endlabel;
          Label label1]@
       (p_instruction inst1)
       @[ Jump endlabel;
          Label endlabel] 
    | Pp.VarExpr (v,b) -> ignore (declarevar v b);[];
    | Pp.ProcDefinition procdef -> declareproc procdef; []; 
    | Pp.Continue -> begin match (!lastcontinue) with
        | None -> incr transerrors;Printf.fprintf stderr "Error : Continue not allowed %s\n" (Pp.pos_string (fst p));[]; 
        | Some label1 -> [Jump label1]
      end    
    | Pp.Break -> begin match (!lastbreak) with
        | None -> incr transerrors;Printf.fprintf stderr "Error : Break not allowed %s\n" (Pp.pos_string (fst p));[]; 
        | Some label1 -> [Jump label1]
      end
    | Pp.Return expr -> begin
        match (expr,!procresult) with
            (Pp.NoneExpr,None)    -> [Return (!currentfunc,Temp [])];
          | (Pp.NoneExpr,Some a)  -> incr transerrors;Printf.fprintf stderr "Error : Return needs an argument %s\n" (Pp.pos_string (fst p)); [];
          | (Pp.Expression a,None)  -> incr transerrors;Printf.fprintf stderr "Error : Return needs no argument %s^n" (Pp.pos_string (fst p)); []; 
          | (Pp.Expression a,Some b) -> [Return (!currentfunc,fst (setcast (p_expression a) (!currentfunc).resulttype (fst p) true) )];    
      end 
    | Pp.NoneInstruction -> [];
and p_expression expr = (*renvoie un Code.exp*Pp.type_expr*)
  match (snd expr) with
    | Pp.Int bigint -> (big_int_to_const bigint,Pp.TConst bigint);
    | Pp.Bool b -> (if b then (Const [1],Pp.TConst (tobig 1)) else (Const [0],Pp.TConst (tobig 0)));
    | Pp.Bin (op,expr1,expr2) ->
	begin match op with
	  | Pp.Plus -> oper_same_type (function ()->Plus) expr1 expr2
	  | Pp.Minus -> oper_same_type (function ()->Minus) expr1 expr2
	  | Pp.And -> oper_same_type (function ()->And) expr1 expr2
	  | Pp.Or -> oper_same_type (function ()->Or) expr1 expr2
	  | Pp.Xor -> oper_same_type (function ()->Xor) expr1 expr2
	  | Pp.L_Or -> oper_bool (function ()->Or) expr1 expr2
          | Pp.L_And -> oper_bool (function ()->And) expr1 expr2
          | Pp.Lt -> oper_sign_homogen (function ()->SLt) (function ()->ULt) expr1 expr2
          | Pp.Le -> oper_sign_homogen (function ()->SLe) (function ()->ULe) expr1 expr2
          | Pp.Gt -> oper_sign_homogen (function ()->SGt) (function ()->UGt) expr1 expr2 
          | Pp.Ge -> oper_sign_homogen (function ()->SGe) (function ()->UGe) expr1 expr2  
          | Pp.Eq -> oper_same_type (function ()->Eq) expr1 expr2  
          | Pp.Ne -> oper_same_type (function ()->Ne) expr1 expr2  
          | Pp.Concat -> 
            let exp1=(p_expression expr1) in 
            let exp2=(p_expression expr2) in begin
               match (snd exp1) with 
                 | Pp.SInteger i -> (Bin (Concat,fst exp1,fst exp2),Pp.SInteger (i+(sizetemps (snd exp2))*8))  
                 | _ -> (Bin (Concat,fst exp1,fst exp2),Pp.UInteger ((sizetemps (snd exp1))*8+(sizetemps (snd exp2))*8))  
  	      end
          | _ -> incr transerrors;Printf.fprintf stderr "Error : operation '%s' not implemented %s\n" (Ast.sbinop op) (Pp.pos_string (fst expr2)); (Temp [0],Pp.Byte);
        end
    | Pp.Un (op,expr2) ->
      begin match op with
        | Pp.Neg -> (
          let exp1=(p_expression expr2) in  
          match (snd exp1) with
            Pp.TConst bigint -> let mbigint=Big_int.minus_big_int bigint in (big_int_to_const mbigint,Pp.TConst mbigint)
            | _ -> (Un (Neg,fst exp1),snd exp1)
          )
        | Pp.Not -> let exp1=(p_expression expr2) in (Un (Not,fst exp1),snd exp1) 
        | Pp.Shr -> (
            let exp1=(p_expression expr2) in (Un (Shr,fst exp1),snd exp1)
          ) 
        | Pp.Shl -> let exp1=(p_expression expr2) in (Un (Shl,fst exp1),snd exp1)
        | Pp.Rol -> let exp1=(p_expression expr2) in (Un (Rol,fst exp1),snd exp1)
        | Pp.Ror -> let exp1=(p_expression expr2) in (Un (Ror,fst exp1),snd exp1)
        | Pp.Ro4 -> let exp1=(p_expression expr2) in (
            match snd exp1 with
              Pp.Byte -> (Un (Ro4,fst exp1),snd exp1)
              | Pp.UInteger 8 -> (Un (Ro4,fst exp1),snd exp1)
              | Pp.SInteger 8 -> ( Printf.fprintf stderr"Warning : This instruction affect variable sign %s\n" (Pp.pos_string (fst expr2));(Un (Ro4,fst exp1),snd exp1) )
              | _ -> incr transerrors;Printf.fprintf stderr "Error : Argument type must be byte,uint8 or sint8 %s\n" (Pp.pos_string (fst expr2));(Temp [0],Pp.Byte);
          )
        | Pp.Cast typ2 -> let exp1=(p_expression expr2) in (setcast exp1 typ2 (fst expr) false)
        | Pp.Extract (i1,i2) -> let exp1=(p_expression expr2) in begin
            if (match (snd exp1) with Pp.Boolean -> true | _ -> false) then (incr transerrors;Printf.fprintf stderr "Error : Can't extract from a boolean expression %s\n" (Pp.pos_string (fst expr2));(Temp [0],Pp.Byte) ) 
            else if (i2<=i1)||(i1<0)||(i1 mod 8<>0)||(i2 mod 8<>7) then (incr transerrors;Printf.fprintf stderr "Error : Invalid extraction range %s\n" (Pp.pos_string (fst expr2));(Temp [0],Pp.Byte) ) 
            else let sz=sizetemps (snd exp1) in 
            let typ=(match (snd exp1) with 
		       | Pp.SInteger i -> if (i<=(i2+1)) then Pp.SInteger (1+i2-i1) else Pp.UInteger (1+i2-i1);
                       | _ -> Pp.UInteger (1+i2-i1)
            ) in (
              if (sz*8<i2) then ( 
                match typ with 
                  | Pp.SInteger i ->  (Un (Extract (i1/8,(i2-7)/8), Un (SExtend (i2-1-sz*8),fst exp1)),typ);
		  | _ ->  (Un (Extract (i1/8,(i2-7)/8), Un (SExtend (i2-1-sz*8),fst exp1)),typ);
              ) else (
                (Un (Extract (i1/8,(i2-7)/8),fst exp1),typ);
	      )
            )
          end
        | Pp.Extractbit (i2) ->  let exp1=(p_expression expr2) in begin
            if (match (snd exp1) with Pp.Boolean -> true | _ -> false) then (incr transerrors;Printf.fprintf stderr "Error : Can't extract from a boolean expression %s\n" (Pp.pos_string (fst expr2));(Temp [0],Pp.Byte) ) 
            else if (i2<0) then (incr transerrors;Printf.fprintf stderr "Error : Invalid bit number %s\n" (Pp.pos_string (fst expr2));(Temp [0],Pp.Byte) ) 
            else let sz=sizetemps (snd exp1) in (
              if (sz*8<i2) then ( 
                match (snd exp1) with 
                  | Pp.SInteger i ->  (Un (Extractbit (i-1),fst exp1),Pp.Boolean);
		  | _ ->  (Const [0],Pp.Boolean);
              ) else (
                (Un (Extractbit i2,fst exp1),Pp.Boolean);
	      )
            )
          end
        | Pp.SAssign sreg -> let exp1=(p_expression expr2) in 
          begin match sreg with
            | Pp.ERA    -> (Un (SAssign Codegen.regERA,fst exp1) ,snd exp1);
            | Pp.ERB    -> (Un (SAssign Codegen.regERB,fst exp1) ,snd exp1);
            | Pp.ERC    -> (Un (SAssign Codegen.regERC,fst exp1) ,snd exp1);
            | Pp.OPTION -> (Un (SAssign Codegen.regOPTION,fst exp1) ,snd exp1); 
          end 
      end     
    | Pp.Get id -> let var=getvar id in (Temp var.temps,var.typ)
    | Pp.Function_call (id,explist)-> (
        let proc=getproc id in       
        let resulttype = (match proc.procdef.Pp.result with None -> Pp.Void | Some a ->a  ) in
        try (let args=List.fold_left2 ( fun lst pexp (epos,typ)-> lst@[
                                                                       fst (setcast (p_expression pexp) typ (fst pexp) true)  ];) 
                        [] explist proc.procdef.Pp.arguments
             in begin
               (Call (proc.funct,args), resulttype );        
             end 
        ) with invalid_arg -> incr transerrors;Printf.fprintf stderr "Error : Function prototype is %s %s\n" (Ast.prototype proc.procdef.Pp.name proc.procdef.Pp.arguments) (Pp.pos_string (fst expr));nothing_of_type resulttype;
      ) 
    | Pp.ESet (id,expr) -> let var=getvar id in
      begin 
       ( Bin (Assign , Temp var.temps, fst (setcast (p_expression expr) var.typ (fst expr) true) ) , var.typ) end;
    | Pp.Incr (a,b,id) -> (
        if (a=0) then
        begin
          let value=(if b=1 then 1 else 255) in
          let var=getvar id in
          let temp=createtemp var.typ in
          (Bin (Comma, (* temp=a;a=a+1;temp *)
            Bin (Comma,
              Bin(Assign, Temp temp.temps, Temp var.temps),
              Bin(Assign, Temp var.temps, 
                Bin(Plus,Temp var.temps,
                  Const ( 
                    let code=ref [] in 
                    (for i=1 to (sizetemps var.typ) do code:=value::(!code); done;!code)
                  ) 
                )
              )
            ), 
          Temp temp.temps ), var.typ)  
        end else begin
          let value=(if b=1 then 1 else 255) in
          let var=getvar id in
          (Bin (Assign, Temp var.temps, 
            Bin (Plus, Temp var.temps,
              Const ( 
                let code=ref [] in 
                (for i=1 to (sizetemps var.typ) do code:=value::(!code); done;!code)
              )
            )
          ),var.typ);
        end;  
      )
    | Pp.Ifte (exp1,exp2,exp3) -> 
       let expr1=setcast (p_expression exp1) Pp.Boolean (fst exp1) true in
       let expr2=p_expression exp2 in
       let expr3=p_expression exp3 in 
       let typ=get_compatible_type (snd expr2) (snd expr3) (fst exp2) in 
       let expr2b=setcast (expr2) typ (fst exp2) true in
       let expr3b=setcast (expr3) typ (fst exp3) true in (Ifte (fst expr1,fst expr2b,fst expr3b,sizetemps typ),typ) 
;; 

let createreg name temp=
  createvar {vid=0;vname=Lexer.int_of_symbol name;typ=Pp.Byte;static=false;temps=[temp]} (Pp.posdummy())

let trans_program p =
  transerrors:=0;
  createreg "RA" Codegen.regRA;
  createreg "RB" Codegen.regRB;
  createreg "RC" Codegen.regRC; 
  createreg "M"  Codegen.regM; 
  let firstlabel = Gen.named_label "_main" in
  let infiniteloop = Gen.new_label () in
  let myfunc= { code = [] ; 
                name = firstlabel; 
                args = [];   (* a completer *)
                result = []; (* a completer *)
                resulttype = Pp.Byte; 
                statics=[];  (* a completer*)
                uses=[];
                defs=[];
                usefuncs = [];
                translating =false;
                translated=false; 
	        depth=0;
              } in                       
  let proccode= ((Label firstlabel)::(Exp (Bin (Assign, Temp [Codegen.regOPTION], Const [0x40])))
		 ::(List.fold_left add_p_instruction [] p)) in               
  begin
    if ((proc_can_define (Lexer.int_of_symbol "main"))=(-1)) then begin
      myfunc.usefuncs <- [];
      myfunc.code <- proccode;
      incr transerrors;Printf.fprintf stderr "Error : main not defined\n";
    end else begin
      let mainfunc=(getproc (Lexer.getpos (),Lexer.int_of_symbol "main")).funct in
      begin
        myfunc.usefuncs <- [mainfunc]; 
        myfunc.code <- proccode@[Exp (Call (mainfunc,[]) );Jump infiniteloop;Label infiniteloop;Jump infiniteloop];
      end;
    end;
    flush stderr;
    flush stdout;
    if (!transerrors>0) then exit(1); 
    { main=myfunc; funcs=(!funclist) };
  end;
;;

